﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using OpenRiaServices.DomainServices.Client.Test.Services;
using DataTests.AdventureWorks.LTS;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading;
using System.Diagnostics;

namespace OpenRiaServices.DomainServices.Client.Test
{
    /// <summary>
    /// Wrapper class for creating new isolated test database instances.
    /// </summary>
    public class TestDatabase : IDisposable
    {
        private string databaseName;
        private bool isInitialized;
        private bool isInitializing;

        public TestDatabase(string databaseName) 
        {
            this.databaseName = databaseName;
        }

        /// <summary>
        /// Idempotent initialization method. Once the database has been initialized
        /// subsequent calls are ignored
        /// </summary>
        public void Initialize()
        {
            if (!isInitializing && !IsInitialized)
            {
                // call the TestServices web service method to create the database 
                TestServicesClient ts = new TestServicesClient();
                SetAddress(ts);
                ts.CreateNewDatabaseCompleted += delegate(object sender, AsyncCompletedEventArgs e)
                {
                    if (e.Error != null)
                    {
                        throw e.Error;
                    }
                    isInitialized = true;
                    isInitializing = false;
                };
                ts.CreateNewDatabaseAsync(databaseName);
                isInitializing = true;
            }
        }

        internal static void SetAddress(TestServicesClient ts)
        {
#if SILVERLIGHT
            ts.Endpoint.Address = new System.ServiceModel.EndpointAddress(
                new Uri(
                    new Uri(
                        System.Windows.Browser.HtmlPage.Document.DocumentUri.GetComponents(UriComponents.SchemeAndServer, UriFormat.Unescaped),
                        UriKind.Absolute),
                    ts.Endpoint.Address.Uri.GetComponents(UriComponents.PathAndQuery, UriFormat.Unescaped)));
#endif
        }

        /// <summary>
        /// Returns true once the async initialization has completed
        /// </summary>
        public bool IsInitialized
        {
            get
            {
                return isInitialized;
            }
        }

        #region IDisposable Members

        void IDisposable.Dispose()
        {
            if (isInitialized)
            {
                TestServicesClient ts = new TestServicesClient();
                SetAddress(ts);
                isInitialized = false;
                ts.ReleaseNewDatabaseAsync(databaseName);    
            }
        }
        #endregion
    }

    public static class TestHelperMethods {
        /// <summary>
        /// Perform a submit on the specified context using the DomainClient directy. This bypasses all context operations,
        /// validation, etc.
        /// </summary>
        /// <param name="ctxt">The context to submit on</param>
        /// <param name="callback">The callback to execute when the submit completes</param>
        public static void SubmitDirect(DomainContext ctxt, Action<SubmitCompletedResult> callback)
        {
            EntityChangeSet cs = ctxt.EntityContainer.GetChanges();
            Debug.Assert(!cs.IsEmpty, "No changes to submit!");

            ctxt.DomainClient.BeginSubmit(cs,
                delegate(IAsyncResult asyncResult)
                {
                    SubmitCompletedResult submitResults = ctxt.DomainClient.EndSubmit(asyncResult);

                    SynchronizationContext syncContext = SynchronizationContext.Current ?? new SynchronizationContext();
                    syncContext.Post(
                        delegate
                        {
                            callback(submitResults);
                        },
                        null);
                }, null);
        }

        /// <summary>
        /// Method used to turn property level validation on or off for an entity. This allows invalid
        /// data to be pumped into the entity, which can be useful in testing.
        /// </summary>
        /// <param name="entity">The entity</param>
        /// <param name="enable">True to turn validation on, false to turn it off.</param>
        public static void EnableValidation(Entity entity, bool enable)
        {
            if (enable)
            {
                entity.OnDeserialized(new System.Runtime.Serialization.StreamingContext());
            }
            else
            {
                entity.OnDeserializing(new System.Runtime.Serialization.StreamingContext());
            }
        }

        /// <summary>
        /// Method used to turn property level validation on or off for a complex object. This allows invalid
        /// data to be pumped into the instance, which can be useful in testing.
        /// </summary>
        /// <param name="entity">The complex object</param>
        /// <param name="enable">True to turn validation on, false to turn it off.</param>
        public static void EnableValidation(ComplexObject complexObject, bool enable)
        {
            if (enable)
            {
                complexObject.OnDeserialized(new System.Runtime.Serialization.StreamingContext());
            }
            else
            {
                complexObject.OnDeserializing(new System.Runtime.Serialization.StreamingContext());
            }
        }

        public static void DefaultOperationAction<T>(T operation) where T : OperationBase
        {
            if (operation.HasError)
            {
                operation.MarkErrorAsHandled();
            }
        }

        public static void AssertOperationSuccess(OperationBase operation)
        {
            string errorMsg = string.Empty;
            if(operation.HasError)
            {
                errorMsg = operation.Error.Message + " : " + operation.Error.StackTrace;
            }
            Assert.IsFalse(operation.HasError, errorMsg);
        }

        public static T CloneEntity<T>(T entity) where T : Entity, new() {
            T clone = new T();
            clone.ApplyState(entity.ExtractState());
            return clone;
        }

        /// <summary>
        /// Returns true if the current entity state equals the specified expected state
        /// </summary>
        public static bool VerifyEntityState(IDictionary<string, object> expectedState, IDictionary<string, object> currentState)
        {
            foreach (var item in currentState) {
                object originalValue = expectedState[item.Key];
                object currentValue = item.Value;
                if (originalValue == null && currentValue == null) {
                    continue;
                }
                if ((originalValue == null && currentValue != null) ||
                    (currentValue == null && originalValue != null)) {
                    return false;
                }
                if (!currentValue.Equals(originalValue)) {
                    return false;
                }
            }
            return true;
        }
    }

    /// <summary>
    /// Class that provides/caches database data without performing any remote queries.
    /// </summary>
    public static class BaselineTestData {
        private static IEnumerable<Product> products = null;

        public static IEnumerable<Product> Products {
            get {
                if (products == null) {
                    products = LoadTestData<Product>(prodData).ToList();
                }
                return products;
            }
        }

        private static IEnumerable<T> LoadTestData<T>(string data) {
            try {
                Type dataType = null;
                // the different fx serializers are deserializing into different types
#if SILVERLIGHT
                dataType = typeof(Dictionary<string, List<object>>);
#else
                dataType = typeof(Dictionary<string, object[]>);
#endif
                System.Runtime.Serialization.Json.DataContractJsonSerializer ser =
                    new System.Runtime.Serialization.Json.DataContractJsonSerializer(dataType, new Type[] { typeof(T) });
                System.Text.UnicodeEncoding encoding = new System.Text.UnicodeEncoding();
                System.IO.MemoryStream ms = new System.IO.MemoryStream(encoding.GetBytes(data));
                object deserializedData = (((System.Collections.IDictionary)ser.ReadObject(ms))[typeof(T).Name]);
                return ((IEnumerable<object>)deserializedData).Cast<T>();
            }
            catch (Exception e) {
                string msg = e.Message;
                throw e;
            }
        }

        /// <summary>
        /// TODO : currently storing raw JSON here for deserialization until I come up with a better
        /// solution for reading data files from both desktop and silverlight. Would be better to read
        /// this data from a file, or include this data as a compiled resource.
        /// </summary>
        #region Raw Data
        static string prodData = "[{\"Key\":\"Product\",\"Value\":[{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":1}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Adjustable Race\",\"ProductID\":1,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"AR-5381\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"694215b7-08f7-4c0d-acb1-d734ba44c0c8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":2}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Bearing Ball\",\"ProductID\":2,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"BA-8327\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"58ae3c20-4f3a-4749-a7d4-d568806cc537\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":3}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"BB Ball Bearing\",\"ProductID\":3,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"BE-2349\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9c21aed2-5bfa-4f18-bcb8-f11638dc2e4e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":4}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Headset Ball Bearings\",\"ProductID\":4,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"BE-2908\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ecfed6cb-51ff-49b5-b06c-7d8ac834db8b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":316}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Blade\",\"ProductID\":316,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"BL-2036\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e73e9750-603b-4131-89f5-3dd15ed5ff80\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":317}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Crankarm\",\"ProductID\":317,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CA-5965\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3c9d10b7-a6b2-4774-9963-c19dcee72fea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":318}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Crankarm\",\"ProductID\":318,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CA-6738\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"eabb9a92-fa07-4eab-8955-f0517b4a4ca7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":319}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Crankarm\",\"ProductID\":319,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CA-7457\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7d3fd384-4f29-484b-86fa-4206e276fe58\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":320}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Chainring Bolts\",\"ProductID\":320,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CB-2903\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7be38e48-b7d6-4486-888e-f53c26735101\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":321}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Chainring Nut\",\"ProductID\":321,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CN-6137\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3314b1d7-ef69-4431-b6dd-dc75268bd5df\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":322}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Chainring\",\"ProductID\":322,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CR-7833\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f0ac2c4d-1a1f-4e3c-b4d9-68aea0ec1ce4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":323}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Crown Race\",\"ProductID\":323,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CR-9981\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"51a32ca6-65a1-4c31-af2b-d9e4f5d631d4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":324}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Chain Stays\",\"ProductID\":324,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"CS-2812\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fe0678ed-aef2-4c58-a450-8151cc24ddd8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":325}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Decal 1\",\"ProductID\":325,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"DC-8732\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"05ce123c-a402-478e-ae9b-75d7727aeaad\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":326}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Decal 2\",\"ProductID\":326,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"DC-9824\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a56851f9-1cd7-4e2f-8779-2e773e1b5209\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":327}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Down Tube\",\"ProductID\":327,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"DT-2377\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1dad47dd-e259-42b8-b8b4-15a0b7d21b2f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":328}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain End Caps\",\"ProductID\":328,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"EC-M092\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6070b1ea-59b7-4f8b-950f-2be07d00449d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":329}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road End Caps\",\"ProductID\":329,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"EC-R098\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"88399d13-719e-4545-81d6-f0650f372fa2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":330}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring End Caps\",\"ProductID\":330,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"EC-T209\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6903ce24-d0ce-4191-9198-4231de37a929\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":331}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Fork End\",\"ProductID\":331,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FE-3760\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"c91d602e-da52-43d2-bd7e-eb110a9392b9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":332}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Freewheel\",\"ProductID\":332,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FH-2981\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d864879a-e8b1-4f7b-bafa-1f136089c2c8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":341}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 1\",\"ProductID\":341,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-1000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a3f2fa3a-22e1-43d8-a131-a9b89c32d8ea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":342}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 6\",\"ProductID\":342,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-1200\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"331addec-e9b9-4a7e-9324-42069c2dcdc4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":343}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 2\",\"ProductID\":343,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-1400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"84a3473e-ae26-4a21-81b9-60bb418a79b2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":344}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 9\",\"ProductID\":344,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-3400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0ae4ce60-5242-48f5-ada1-3013ff45f969\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":345}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 4\",\"ProductID\":345,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-3800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2c1c58b4-234c-4b3a-8c8e-84524ac05eea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":346}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 3\",\"ProductID\":346,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-5160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"590c2c3f-a8b6-42b5-9412-d655e37f0eae\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":347}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 8\",\"ProductID\":347,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-5800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1b73f5fe-ab85-49fc-99ad-0500cebda91d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":348}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 5\",\"ProductID\":348,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-7160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d182cf18-4ddf-429b-a0df-de1cfc92622d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":349}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Flat Washer 7\",\"ProductID\":349,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FW-9160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7e55f64d-ea3c-45ff-be72-f7f7b9d61a79\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":350}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Fork Crown\",\"ProductID\":350,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FC-3654\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1cbfa85b-5c9b-4b58-9c17-95238215d926\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":351}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Front Derailleur Cage\",\"ProductID\":351,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FC-3982\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"01c901e3-4323-48ed-ab9e-9bfda28bdef6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":352}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Front Derailleur Linkage\",\"ProductID\":352,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"FL-2301\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"88ed2e08-e775-4915-b506-831600b773fd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":355}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Guide Pulley\",\"ProductID\":355,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"GP-0982\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6a89205b-90c3-4997-8c63-bc6a5ebc752d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":356}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Grip Tape\",\"ProductID\":356,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"GT-0820\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"32c82181-1969-4660-ae04-a02d51994ec1\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":357}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Grip Tape\",\"ProductID\":357,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"GT-1209\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"09343e22-2494-4279-9f32-5d961a0fbfea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":358}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Grip Tape\",\"ProductID\":358,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"GT-2908\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8e5b2bf7-81dd-4454-b75e-d9ae2a6fbd26\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":359}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 9\",\"ProductID\":359,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-1213\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a63aff3c-4143-4016-bc99-d3f80ec1ade5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":360}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 10\",\"ProductID\":360,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-1220\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a1ae0c6d-92a5-4fda-b33b-1301e83efe5b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":361}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 1\",\"ProductID\":361,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-1420\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e33e8c4c-282a-4d1f-91e7-e9969cf7134f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":362}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 2\",\"ProductID\":362,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-1428\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a992684f-4642-4856-a817-2c0740ae8c55\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":363}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 15\",\"ProductID\":363,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-3410\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b9d9a30d-cb07-422d-a312-f0535575337e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":364}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 16\",\"ProductID\":364,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-3416\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0ec8f653-24c9-41b6-86f5-39c1789db580\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":365}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 5\",\"ProductID\":365,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-3816\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"94506c9d-5991-46a7-82ea-00e05d8d9b9c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":366}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 6\",\"ProductID\":366,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-3824\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"71e984c6-1d11-4cf2-baee-6c8df494bdad\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":367}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 3\",\"ProductID\":367,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-5161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"efc09cdb-ecd5-4db5-9e27-277dda6d7f50\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":368}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 4\",\"ProductID\":368,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-5162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0a0c93aa-d06c-48aa-99eb-20f2c8a6d6c4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":369}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 13\",\"ProductID\":369,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-5811\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a2728648-9517-4dec-8606-d7d98ecd1e91\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":370}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 14\",\"ProductID\":370,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-5818\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0a7ad37c-3696-4844-8633-9fddcd5fcefc\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":371}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 7\",\"ProductID\":371,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-7161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"acbb1de1-680c-4034-a8c5-3c6b01e57aa7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":372}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 8\",\"ProductID\":372,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-7162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a0da8f8f-45fb-4e62-ab14-aa229087de1e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":373}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 12\",\"ProductID\":373,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-9080\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"db99cbcd-f18d-4979-8dcf-1012f3b1cb15\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":374}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Hex Nut 11\",\"ProductID\":374,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HJ-9161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ec2e54f5-9d07-4c26-b969-40f835395be3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":375}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 5\",\"ProductID\":375,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1024\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f2f3a14c-df15-4957-966d-55e5fcad1161\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":376}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 6\",\"ProductID\":376,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1032\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e73e44dd-f0b7-45d4-9066-e49f1b1fe7d0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":377}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 16\",\"ProductID\":377,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1213\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2b2a5641-bffe-4079-b9f0-8bf355bc3996\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":378}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 17\",\"ProductID\":378,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1220\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f70bbecd-5be7-4ee9-a9e7-15786e622ba9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":379}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 7\",\"ProductID\":379,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1224\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ba3631d1-33d6-4a2f-8413-758bfda6f9c2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":380}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 8\",\"ProductID\":380,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1420\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b478b33a-1fd5-4db6-b99a-eb3b2a7c1623\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":381}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 9\",\"ProductID\":381,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-1428\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"da46d979-59df-456d-b5ae-e7e89fc589df\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":382}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 22\",\"ProductID\":382,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-3410\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2f457404-197d-4ddf-9868-a3aad1b32d6b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":383}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 23\",\"ProductID\":383,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-3416\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5f02449e-96e5-4fc8-ade0-8a9a7f533624\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":384}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 12\",\"ProductID\":384,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-3816\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"32c97696-7c3d-4793-a54b-3d73200badbc\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":385}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 13\",\"ProductID\":385,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-3824\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8f0902d0-274d-4a4b-8fde-e37f53b2ab29\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":386}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 1\",\"ProductID\":386,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-4402\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"57456f8c-cb78-45ec-b9b8-21a9be5c12f5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":387}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 10\",\"ProductID\":387,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-5161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"955811b4-2f17-48f0-a8b4-0c96cba4aa6d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":388}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 11\",\"ProductID\":388,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-5162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7d4bad17-374f-4281-9ae5-49abc3fe585d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":389}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 2\",\"ProductID\":389,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-5400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"17c4b7ba-8574-4ec7-bd3b-7a51aba61f69\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":390}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 20\",\"ProductID\":390,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-5811\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b4e990b7-b3f7-4f97-8f98-ce373833adb4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":391}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 21\",\"ProductID\":391,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-5818\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"433ee702-e028-4630-895c-8cdbd0f1fd89\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":392}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 3\",\"ProductID\":392,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-6320\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ee76fae0-161e-409c-a6f5-837b5b8b344d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":393}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 14\",\"ProductID\":393,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-7161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"eb10b88c-5351-4c06-af51-116baa48a2c8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":394}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 15\",\"ProductID\":394,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-7162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"35e28755-e8f0-47be-a8be-a20836dbe28d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":395}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 4\",\"ProductID\":395,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-8320\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"69ab8d5e-6101-4203-81b1-794e923782ea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":396}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 18\",\"ProductID\":396,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-9161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"39d42384-66f6-4ccd-b471-0589fc3fc576\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":397}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hex Nut 19\",\"ProductID\":397,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HN-9168\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b63f827e-9055-4678-9e90-4ffd8d06d4d4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":398}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Handlebar Tube\",\"ProductID\":398,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HT-2981\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9f88c58e-5cfa-46c9-8994-da4f3ffe92ed\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":399}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Head Tube\",\"ProductID\":399,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HT-8019\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b047c048-b4fb-4f80-94bc-c5fc00a6f77f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":400}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Hub\",\"ProductID\":400,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HU-6280\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ab332dda-dda5-44ad-8c50-34ffaceffa8e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":401}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Hub\",\"ProductID\":401,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"HU-8998\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d59403a3-d19e-4803-bda2-b436a6416fda\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":402}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Keyed Washer\",\"ProductID\":402,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"KW-4091\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"43024784-2741-4cab-a6dc-8050ce507f2e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":403}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 3\",\"ProductID\":403,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-1000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"763412f0-6d53-43e2-b371-dcbed69f5e98\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":404}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 4\",\"ProductID\":404,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-1200\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"89b6e84f-5c08-4cd9-b803-01f2ce24e417\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":405}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 9\",\"ProductID\":405,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-1201\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3330a721-e8cb-4e67-8d27-300db68c2395\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":406}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 5\",\"ProductID\":406,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-1400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"39314098-768b-49f9-a777-af2e3bb06b17\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":407}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 7\",\"ProductID\":407,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-3800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"93f7bc73-bd22-45a0-9f2e-a11932342e6b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":408}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 6\",\"ProductID\":408,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-5160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"dc5f4cb0-a599-42cd-a96f-e9f01bc1dacc\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":409}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 1\",\"ProductID\":409,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-6000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"802b7687-bc74-43f8-98ae-2112166faca7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":410}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 8\",\"ProductID\":410,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-7160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"02c48826-21ad-41f3-85a6-bc9a85a4dce4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":411}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"External Lock Washer 2\",\"ProductID\":411,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LE-8000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"502a2a3d-cd72-43ad-8fb6-77505187edf4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":412}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 3\",\"ProductID\":412,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-1000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f1168c45-e4d2-4c37-adad-b76eaf402b71\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":413}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 4\",\"ProductID\":413,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-1200\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7f7413bb-bad2-47e4-9bc4-d98b194be35d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":414}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 9\",\"ProductID\":414,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-1201\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"4f040109-8332-4fcf-8084-57e6d8c49283\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":415}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 5\",\"ProductID\":415,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-1400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0c02f2cc-bdb4-4794-a4f9-0eb33f7545c2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":416}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 7\",\"ProductID\":416,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-3800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3c2ac5bc-3f49-4fa4-8340-bc4cda983f46\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":417}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 6\",\"ProductID\":417,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-5160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7f175dfe-1669-4ee9-8eeb-7b55fce9961c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":418}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 10\",\"ProductID\":418,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-5800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"c8323eec-bdb2-4933-b3c6-24287638ad56\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":419}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 1\",\"ProductID\":419,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-6000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"71f8232d-2b59-41ac-99a1-f5ea197671b5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":420}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 8\",\"ProductID\":420,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-7160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e2f03586-02e8-4cd9-a342-1a8d65d393bd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":421}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Internal Lock Washer 2\",\"ProductID\":421,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LI-8000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"97741a88-92a1-4e72-b0aa-bcb198a1c378\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":422}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 9\",\"ProductID\":422,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-1213\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7da2613b-3347-4072-a1dc-943ada161d7f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":423}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 10\",\"ProductID\":423,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-1220\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a88f15be-2719-4741-93a4-2d96ef3712eb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":424}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 1\",\"ProductID\":424,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-1420\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"48461e5d-d58a-47e5-8ba3-ce4430fca611\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":425}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 2\",\"ProductID\":425,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-1428\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"36187eb6-ad84-47b7-a55e-2941d3435fe2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":426}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 15\",\"ProductID\":426,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-3410\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"99215648-afe8-4556-bc80-b6c62ae74fc8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":427}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 16\",\"ProductID\":427,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-3416\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b4fc4c32-049c-417f-bbb0-f19cdfd64252\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":428}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 5\",\"ProductID\":428,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-3816\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a57b7915-2e65-49de-87ba-acd007c55adb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":429}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 6\",\"ProductID\":429,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-3824\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5abd940c-d61f-4108-8312-0ea97e469613\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":430}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 3\",\"ProductID\":430,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-5161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9883496f-4785-4bfc-8af3-c357347b9f89\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":431}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 4\",\"ProductID\":431,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-5162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8c5ee683-d93c-4f25-9454-22faa4c31365\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":432}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 13\",\"ProductID\":432,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-5811\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"38e4a447-3d3c-4087-abad-97f006525b52\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":433}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 14\",\"ProductID\":433,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-5818\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"dce24b6c-76d8-4934-a4f6-c934364943ea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":434}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 7\",\"ProductID\":434,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-7161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"344ad07c-cca5-4374-a3f3-8a7e0a1d9916\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":435}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 8\",\"ProductID\":435,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-7162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b2508cf2-c64f-493d-9db4-0d6601af1f73\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":436}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 12\",\"ProductID\":436,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-9080\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5d3e589f-4584-406b-b9cc-3c8c060cb9a5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":437}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Thin-Jam Lock Nut 11\",\"ProductID\":437,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LJ-9161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"64169a28-161c-4737-b724-f42ffd99de80\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":438}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 5\",\"ProductID\":438,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1024\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"eb4c1d34-4816-4130-bb30-07b4de4072b6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":439}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 6\",\"ProductID\":439,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1032\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"98ccbb38-4683-486e-bbfe-cbbe4ea63c03\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":440}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 16\",\"ProductID\":440,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1213\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"bbfd88f8-28c5-44ee-b625-52e882393dfc\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":441}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 17\",\"ProductID\":441,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1220\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"92dc4ba8-a052-45df-83ec-226f050f876b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":442}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 7\",\"ProductID\":442,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1224\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d583d825-c707-4529-b6f2-abffa21b81ec\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":443}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 8\",\"ProductID\":443,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1420\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e91c3dc2-c99b-4d01-8108-5dde3c87830a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":444}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 9\",\"ProductID\":444,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-1428\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"06534a20-4c00-4824-8bba-b4e3a5724d93\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":445}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 22\",\"ProductID\":445,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-3410\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1e4fa4ec-367e-4d8d-88b4-6cd34d1cfb89\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":446}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 23\",\"ProductID\":446,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-3416\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"afa814c8-8ec8-49db-9fee-a291197a8fe9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":447}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 12\",\"ProductID\":447,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-3816\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9a751f85-7130-4562-9f22-db9cab6e5bbc\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":448}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 13\",\"ProductID\":448,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-3824\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"06be8347-45c1-4c40-afcb-6ab34ad135fb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":449}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 1\",\"ProductID\":449,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-4400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1dc29704-e0e0-4ef5-b581-4a524730c448\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":450}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 10\",\"ProductID\":450,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-5161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"612c26c7-6018-4050-b628-8b2d2e6841fa\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":451}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 11\",\"ProductID\":451,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-5162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5bcc4558-6c16-48f1-92f0-fd2eefb17306\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":452}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 2\",\"ProductID\":452,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-5400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"53ad147d-c16d-4a8c-b086-625a31405574\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":453}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 20\",\"ProductID\":453,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-5811\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2749030e-49b7-4b24-8d47-fbcf194aba38\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":454}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 21\",\"ProductID\":454,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-5818\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e10a7b34-87f5-42cd-88b3-27a3d8e16b18\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":455}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 3\",\"ProductID\":455,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-6320\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"aa5071eb-2145-4d08-9d33-b9d2ba9e1493\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":456}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 14\",\"ProductID\":456,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-7161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1e96b03d-dc07-4a98-bc24-bf5b24c393e5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":457}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 15\",\"ProductID\":457,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-7162\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ce04de2b-4eca-4203-a108-b7d92ff0e96e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":458}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 4\",\"ProductID\":458,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-8320\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"41bd9389-8b22-4a35-9a2c-233d39ada7ea\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":459}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 19\",\"ProductID\":459,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-9080\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5986504b-22a0-4e74-a137-c7cf99a8216f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":460}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Nut 18\",\"ProductID\":460,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LN-9161\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a420963f-92fd-4cd4-8531-6926e4162c41\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":461}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Ring\",\"ProductID\":461,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LR-2398\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"aeca59da-b61c-4976-8316-97e14cd4eff1\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":462}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lower Head Race\",\"ProductID\":462,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LR-8520\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"dbb962bf-0603-4095-959b-5ba9c74fbd32\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":463}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 4\",\"ProductID\":463,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-1000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a3ee3bc5-73c5-45f3-a952-9991d038dd9c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":464}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 5\",\"ProductID\":464,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-1200\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ecaed08d-2cf5-4d81-a4ed-8369e25af227\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":465}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 10\",\"ProductID\":465,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-1201\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a2212bab-af58-41a5-a659-a6141c8967ca\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":466}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 6\",\"ProductID\":466,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-1400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9092f2e1-3637-4669-8565-55404a55750e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":467}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 13\",\"ProductID\":467,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-3400\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3cb31f4a-c61c-408c-be1e-48bee412e511\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":468}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 8\",\"ProductID\":468,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-3800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"900d26e6-21a0-427d-b43c-173f6dcb2045\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":469}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 1\",\"ProductID\":469,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-4000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5402ea37-29df-47ff-9fc7-867d60594c48\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":470}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 7\",\"ProductID\":470,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-5160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"99357255-e66b-458c-ab2e-6f68ef5452d2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":471}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 12\",\"ProductID\":471,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-5800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7bc9d58e-3e62-481f-8343-beb0883b3ecf\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":472}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 2\",\"ProductID\":472,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-6000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5f201424-9e6a-4f8d-9c2c-30777e27d64f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":473}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 9\",\"ProductID\":473,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-7160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f9426fb2-1e68-464e-bf32-615026e0249e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":474}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 3\",\"ProductID\":474,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-8000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ac007b7f-73b7-4623-8150-02444c5ec023\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":475}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Lock Washer 11\",\"ProductID\":475,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"LW-9160\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"639d8448-b427-47b1-9e5b-0e5090a27632\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":476}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Angle\",\"ProductID\":476,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MA-7075\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e876e239-7ec2-45c8-ba4b-b9ceacb379a6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":477}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Bar 1\",\"ProductID\":477,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MB-2024\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8b5429ce-7876-44b3-9332-baf78a238b36\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":478}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Bar 2\",\"ProductID\":478,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MB-6061\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2a14f60e-3827-49ba-af13-466dbc30c5ba\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":479}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Plate 2\",\"ProductID\":479,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MP-2066\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0773a2c9-f47f-429e-814a-25b2e08c128a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":480}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Plate 1\",\"ProductID\":480,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MP-2503\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"242389be-dde0-42a1-85d9-f99fdc981336\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":481}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Plate 3\",\"ProductID\":481,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MP-4960\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8b7e90e5-7785-455e-bc7c-e962f18c6848\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":482}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 2\",\"ProductID\":482,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-0253\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8bb96dfb-23aa-4877-9c5e-866bb18facc7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":483}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 3\",\"ProductID\":483,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-1256\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9074e00d-005b-450e-9c92-6667782e8108\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":484}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 7\",\"ProductID\":484,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-1981\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a9539885-0cee-4aa0-9072-8db1d34a16db\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":485}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 4\",\"ProductID\":485,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-2259\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3cb3cf7d-ab8e-44a3-b7e9-73149f5ec29f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":486}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 5\",\"ProductID\":486,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-2341\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2a2c555d-328d-4299-bd83-591d0762df62\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":487}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 6\",\"ProductID\":487,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-2348\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"64844011-a1c3-4f8f-9caa-9c8d214ecc12\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":488}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Sheet 1\",\"ProductID\":488,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MS-6061\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3b2febc6-c76c-4a56-9cf7-8af5b76e24ee\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":489}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Metal Tread Plate\",\"ProductID\":489,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"MT-1000\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d2177b6c-3352-43f0-9a41-719754dd280c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":490}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Nipple\",\"ProductID\":490,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"NI-4127\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"88310f73-ab0a-41a2-8597-936f192b7d12\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":491}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Nipple\",\"ProductID\":491,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"NI-9522\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"88a7b897-6ff5-4ca2-b68a-6ea0e86f92b9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":492}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Paint - Black\",\"ProductID\":492,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"PA-187B\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":45,\"SafetyStockLevel\":60,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"df20e514-3d47-491b-9454-0911ec3f7d29\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":493}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Paint - Red\",\"ProductID\":493,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"PA-361R\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":45,\"SafetyStockLevel\":60,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"4c568357-5d21-4ad4-bb85-bb5519b3b50c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":494}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Paint - Silver\",\"ProductID\":494,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"PA-529S\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":45,\"SafetyStockLevel\":60,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fa81e47d-7333-49c2-809b-308171ca2fb1\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":495}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Paint - Blue\",\"ProductID\":495,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"PA-632U\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":45,\"SafetyStockLevel\":60,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"25a73761-ae90-49d3-8d1d-dd7858db4704\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":496}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Paint - Yellow\",\"ProductID\":496,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"PA-823Y\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":45,\"SafetyStockLevel\":60,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1c8adb43-9fe8-44a6-b949-8af33ce9486e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":497}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Pinch Bolt\",\"ProductID\":497,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"PB-6109\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f1694c24-dfab-4c92-bc66-6e717db24ea8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":504}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Cup-Shaped Race\",\"ProductID\":504,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RA-2345\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"874c800e-334d-4a3c-8d3a-1e872d5b2a1b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":505}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Cone-Shaped Race\",\"ProductID\":505,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RA-7490\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"35ce3995-9dd2-40e2-98b8-275931ac2d76\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":506}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Reflector\",\"ProductID\":506,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RF-9198\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1c850499-38ed-4c2d-8665-7edb6a7ce93d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":507}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Rim\",\"ProductID\":507,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-M464\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":435.00,\"rowguid\":\"b2cc7dfb-783d-4587-88c0-2712a538a5b2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":508}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Rim\",\"ProductID\":508,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-M692\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":450.00,\"rowguid\":\"733fd04d-322f-44f5-beec-f326189d1ce6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":509}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Rim\",\"ProductID\":509,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-M823\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":400.00,\"rowguid\":\"9fa4a3b5-d396-48d4-adfc-b573bc4a800a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":510}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Rim\",\"ProductID\":510,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-R436\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":445.00,\"rowguid\":\"c2770757-b258-4eec-a811-6856faf87437\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":511}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Rim\",\"ProductID\":511,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-R600\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":450.00,\"rowguid\":\"80108059-0002-4253-a805-53a2324c33a4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":512}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Rim\",\"ProductID\":512,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-R800\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":400.00,\"rowguid\":\"cd9b5c44-fb31-4e0f-9905-3b2086966cc5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":513}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring Rim\",\"ProductID\":513,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RM-T801\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":460.00,\"rowguid\":\"4852db13-308a-4893-aafa-390a0dfe9f12\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":514}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":133.3400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Seat Assembly\",\"ProductID\":514,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-M198\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":98.7700,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fcfc0a4f-4563-4e0b-bff4-5ddcfe3a9273\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":515}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":147.1400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Seat Assembly\",\"ProductID\":515,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-M237\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":108.9900,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d3c8ae4c-a1be-448d-bf58-6ecbf36afa0b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":516}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":196.9200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Seat Assembly\",\"ProductID\":516,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-M687\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":145.8700,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9e18adab-b9c7-45b1-bd95-1805ec4f297d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":517}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":133.3400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Seat Assembly\",\"ProductID\":517,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-R127\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":98.7700,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f5a30b8d-f35b-43f2-83a0-f7f6b51f6241\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":518}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":147.1400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Seat Assembly\",\"ProductID\":518,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-R430\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":108.9900,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ad109395-fda9-4c2a-96f1-515ccde3d9f4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":519}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":196.9200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Seat Assembly\",\"ProductID\":519,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-R522\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":145.8700,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7b52ee2a-7100-4a39-a0af-c89012da6ef8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":520}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":133.3400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Seat Assembly\",\"ProductID\":520,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-T467\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":98.7700,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"af3d83ba-4b8e-4072-817f-e6b095a1c879\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":521}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":147.1400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Touring Seat Assembly\",\"ProductID\":521,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-T612\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":108.9900,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"85b9a3de-000c-4351-9494-05796689c216\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":522}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":196.9200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Seat Assembly\",\"ProductID\":522,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SA-T872\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":145.8700,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8c471bca-a735-4087-ad50-90ede0ac1a1b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":523}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Spindle\\/Axle\",\"ProductID\":523,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SD-2342\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d2bd1f55-2cd4-4998-89fa-28ff2e28de2c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":524}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Spindle\\/Axle\",\"ProductID\":524,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SD-9872\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6ce0661d-ba1f-4012-b785-55165b3b241a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":525}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Shell\",\"ProductID\":525,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SH-4562\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ae7bcda7-e836-4f68-9e61-745f27f9aa3e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":526}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Shell\",\"ProductID\":526,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SH-9312\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d215a3ae-aaf2-4cb0-9d20-3758aad078e2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":527}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Spokes\",\"ProductID\":527,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SK-9283\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5aabb729-343b-4084-a235-ccb3da9f29e7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":528}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Seat Lug\",\"ProductID\":528,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SL-0931\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":750,\"SafetyStockLevel\":1000,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"4a898b1e-9a3b-4beb-9873-a7465934051a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":529}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Stem\",\"ProductID\":529,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SM-9087\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1173306e-b616-4c4a-b715-4e0a483ba2b5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":530}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Seat Post\",\"ProductID\":530,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SP-2981\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9b4ceb84-4e84-43f3-b326-9b7f22905363\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":531}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Steerer\",\"ProductID\":531,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SR-2098\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f3b140a1-b139-4bb5-b144-1b7cbbee6c9a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":532}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Seat Stays\",\"ProductID\":532,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"SS-2985\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"20c2c611-dffc-49b5-99cf-d89bdd3a91ce\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":533}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Seat Tube\",\"ProductID\":533,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"ST-9828\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"41f5388b-7253-4002-bcc6-b2a50920d11f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":534}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Top Tube\",\"ProductID\":534,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"TO-2301\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"4c0bad8e-066b-46b8-bfe9-da61539606e8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":535}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Tension Pulley\",\"ProductID\":535,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"TP-0923\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":600,\"SafetyStockLevel\":800,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"13df62b2-8a7b-47d5-9084-f1172c4779e4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":679}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":false,\"ListPrice\":0.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Rear Derailleur Cage\",\"ProductID\":679,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":null},\"ProductNumber\":\"RC-0291\",\"ProductSubcategoryReference\":{\"EntityKey\":null},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":null,\"StandardCost\":0.0000,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"912b03ea-4447-48c8-85da-09b80ab26340\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":680}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Black, 58\",\"ProductID\":680,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92B-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":1059.3100,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.24,\"rowguid\":\"43dd68d6-14a4-461f-9069-55309d90ea7e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":706}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Red, 58\",\"ProductID\":706,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92R-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(896684400000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":1059.3100,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.24,\"rowguid\":\"9540ff17-2712-4c90-a3d1-8ce5568b2462\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":707}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Red\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":34.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Sport-100 Helmet, Red\",\"ProductID\":707,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":33}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HL-U509-R\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":31}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":null,\"StandardCost\":13.0863,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2e1ef41a-c08a-4ff6-8ada-bde58b64a712\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":708}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":34.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Sport-100 Helmet, Black\",\"ProductID\":708,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":33}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HL-U509\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":31}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":null,\"StandardCost\":13.0863,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a25a44fb-c2de-4268-958f-110b8d7621e2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":709}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"White\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":9.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain Bike Socks, M\",\"ProductID\":709,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":18}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SO-B909-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":23}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":3.3963,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"18f95f47-1540-4e02-8f1f-cc1bcb6828d0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":710}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"White\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":9.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain Bike Socks, L\",\"ProductID\":710,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":18}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SO-B909-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":23}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":3.3963,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"161c035e-21b3-4e14-8e44-af508f35d80a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":711}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Blue\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":34.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Sport-100 Helmet, Blue\",\"ProductID\":711,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":33}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HL-U509-B\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":31}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":null,\"StandardCost\":13.0863,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fd7c0858-4179-48c2-865b-abd5dfc7bc1d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":712}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":8.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"AWC Logo Cap\",\"ProductID\":712,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":2}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"CA-1098\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":19}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":null,\"StandardCost\":6.9223,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b9ede243-a6f4-4629-b1d4-ffe1aedc6de7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":713}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":49.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Long-Sleeve Logo Jersey, S\",\"ProductID\":713,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":11}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LJ-0192-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":38.4923,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fd449c82-a259-4fae-8584-6ca0255faf68\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":714}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":49.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Long-Sleeve Logo Jersey, M\",\"ProductID\":714,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":11}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LJ-0192-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":38.4923,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6a290063-a0cf-432a-8110-2ea0fda14308\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":715}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":49.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Long-Sleeve Logo Jersey, L\",\"ProductID\":715,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":11}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LJ-0192-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":38.4923,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"34cf5ef5-c077-4ea0-914a-084814d5cbd5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":716}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":49.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Long-Sleeve Logo Jersey, XL\",\"ProductID\":716,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":11}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LJ-0192-X\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"XL\",\"StandardCost\":38.4923,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6ec47ec9-c041-4dda-b686-2125d539ce9b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":717}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Red, 62\",\"ProductID\":717,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92R-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.30,\"rowguid\":\"052e4f8b-0a2a-46b2-9f42-10febcfae416\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":718}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Red, 44\",\"ProductID\":718,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.12,\"rowguid\":\"a88d3b54-2cae-43f2-8c6e-ea1d97b46a7c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":719}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Red, 48\",\"ProductID\":719,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.16,\"rowguid\":\"07befc9a-7634-402b-b234-d7797733baaf\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":720}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Red, 52\",\"ProductID\":720,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.20,\"rowguid\":\"fcfea68f-310e-4e6e-9f99-bb17d011ebae\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":721}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Red, 56\",\"ProductID\":721,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92R-56\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"56\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.24,\"rowguid\":\"56c85873-4993-41b4-8096-1067cfd7e4bd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":722}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Black, 58\",\"ProductID\":722,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38B-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":204.6251,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.46,\"rowguid\":\"2140f256-f705-4d67-975d-32de03265838\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":723}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Black, 60\",\"ProductID\":723,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38B-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":204.6251,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.48,\"rowguid\":\"aa95e2a5-e7c4-4b74-b1ea-b52ee3b51537\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":724}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Black, 62\",\"ProductID\":724,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38B-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":204.6251,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.50,\"rowguid\":\"5247be33-50bf-4527-8a30-a39aae500a8e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":725}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Red, 44\",\"ProductID\":725,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":187.1571,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.32,\"rowguid\":\"137d319d-44ad-42b2-ab61-60b9ce52b5f2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":726}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Red, 48\",\"ProductID\":726,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":187.1571,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.36,\"rowguid\":\"35213547-275f-4767-805d-c8a4b8e13745\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":727}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Red, 52\",\"ProductID\":727,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":187.1571,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.40,\"rowguid\":\"c455e0b3-d716-419d-abf0-7e03efdd2e26\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":728}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Red, 58\",\"ProductID\":728,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38R-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":187.1571,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.46,\"rowguid\":\"799a56ff-5ad2-41b3-bfac-528b477ad129\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":729}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Red, 60\",\"ProductID\":729,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38R-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":187.1571,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.48,\"rowguid\":\"1784bb14-d1f5-4b24-92da-9127ad179302\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":730}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Red, 62\",\"ProductID\":730,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38R-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":187.1571,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.50,\"rowguid\":\"7e73aa1f-8569-4d87-9f80-ac2e513e0803\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":731}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame - Red, 44\",\"ProductID\":731,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":16}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":352.1394,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.22,\"rowguid\":\"459e041c-3234-409e-b4cd-81728f8a2398\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":732}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame - Red, 48\",\"ProductID\":732,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":16}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":352.1394,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.26,\"rowguid\":\"b673189c-c042-413b-8194-73bc44b0492c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":733}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame - Red, 52\",\"ProductID\":733,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":16}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":352.1394,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.30,\"rowguid\":\"55ea276b-82d8-4ccb-9ab1-9b1b75b15a83\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":734}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame - Red, 58\",\"ProductID\":734,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":16}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72R-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":352.1394,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.36,\"rowguid\":\"df4ce1e2-ba9a-4657-b999-ccfa6c55d9c1\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":735}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame - Red, 60\",\"ProductID\":735,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":16}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72R-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":352.1394,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.38,\"rowguid\":\"b2e48e8c-63a5-469a-ba4c-4f5ebb1104a4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":736}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Black, 44\",\"ProductID\":736,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":204.6251,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.32,\"rowguid\":\"c9967889-f490-4a66-943a-bce432e938d8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":737}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Black, 48\",\"ProductID\":737,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":204.6251,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.36,\"rowguid\":\"3b5f29b6-a441-4ff7-a0fa-fad10e2ceb4c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":738}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":337.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Frame - Black, 52\",\"ProductID\":738,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":9}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R38B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":204.6251,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.40,\"rowguid\":\"18fc5d72-a012-4dc7-bb35-0d01a84d0219\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":739}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1364.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Silver, 42\",\"ProductID\":739,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":747.2002,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.72,\"rowguid\":\"8ae32663-8d6f-457d-8343-5b181fec43a7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":740}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1364.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Silver, 44\",\"ProductID\":740,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94S-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":706.8110,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.76,\"rowguid\":\"1909c60c-c490-411d-b3e6-12ddd7832482\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":741}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1364.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Silver, 48\",\"ProductID\":741,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94S-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":706.8110,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.80,\"rowguid\":\"b181ec1f-ca20-4724-b2eb-15f3e455142e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":742}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1364.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Silver, 46\",\"ProductID\":742,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94S-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":747.2002,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.84,\"rowguid\":\"a189d86e-d923-4336-b13d-a5db6f426540\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":743}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1349.6000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Black, 42\",\"ProductID\":743,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94B-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":739.0410,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.72,\"rowguid\":\"27db28f8-5ab8-4091-b94e-6f1b2d8e7ab0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":744}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1349.6000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Black, 44\",\"ProductID\":744,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":699.0928,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.76,\"rowguid\":\"cb443286-6b25-409f-a10b-1ad4eeb4bd4e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":745}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1349.6000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Black, 48\",\"ProductID\":745,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":699.0928,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.80,\"rowguid\":\"1fee0573-6676-432d-8d6d-41ba9faa5865\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":746}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1349.6000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Black, 46\",\"ProductID\":746,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94B-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":739.0410,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.84,\"rowguid\":\"50abebcb-451e-42b9-8dbb-e5c4a34470e9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":747}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1349.6000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Black, 38\",\"ProductID\":747,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94B-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":739.0410,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.68,\"rowguid\":\"0c548577-3171-4ce2-b9a0-1ed526849de8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":748}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1364.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Frame - Silver, 38\",\"ProductID\":748,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":5}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M94S-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":747.2002,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.68,\"rowguid\":\"f246acaa-a80b-40ec-9208-02edef885129\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":749}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3578.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-150 Red, 62\",\"ProductID\":749,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":25}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R93R-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":2171.2942,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.00,\"rowguid\":\"bc621e1f-2553-4fdc-b22e-5e44a9003569\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":750}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3578.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-150 Red, 44\",\"ProductID\":750,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":25}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R93R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":2171.2942,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":13.77,\"rowguid\":\"c19e1136-5da4-4b40-8758-54a85d7ea494\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":751}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3578.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-150 Red, 48\",\"ProductID\":751,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":25}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R93R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":2171.2942,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":14.13,\"rowguid\":\"d10b7cc1-455e-435b-a08f-ec5b1c5776e9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":752}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3578.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-150 Red, 52\",\"ProductID\":752,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":25}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R93R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":2171.2942,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":14.42,\"rowguid\":\"5e085ba0-3cd5-487f-85bb-79ed1c701f23\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":753}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3578.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-150 Red, 56\",\"ProductID\":753,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":25}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R93R-56\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"56\",\"StandardCost\":2171.2942,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":14.68,\"rowguid\":\"30819b88-f0d3-4e7a-8105-19f6fac2cefb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":754}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1457.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-450 Red, 58\",\"ProductID\":754,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":28}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R68R-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":884.7083,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":17.79,\"rowguid\":\"40d5effa-c0c4-479f-af66-5f1bf8ed3bfb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":755}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1457.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-450 Red, 60\",\"ProductID\":755,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":28}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R68R-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":884.7083,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":17.90,\"rowguid\":\"181a90cb-3678-490e-8418-78f73fb5343d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":756}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1457.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-450 Red, 44\",\"ProductID\":756,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":28}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R68R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":884.7083,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":16.77,\"rowguid\":\"f8b5e26a-3d33-4e39-b500-cc21a133062e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":757}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1457.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-450 Red, 48\",\"ProductID\":757,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":28}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R68R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":884.7083,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":17.13,\"rowguid\":\"c72c9978-0b04-46b3-9de6-948feca1c86e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":758}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1457.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-450 Red, 52\",\"ProductID\":758,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":28}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R68R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":884.7083,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":17.42,\"rowguid\":\"040a4b7d-4060-4507-aa92-7508b434797e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":759}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Red, 58\",\"ProductID\":759,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50R-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.79,\"rowguid\":\"6711d6bc-664f-4890-9f69-af1de321d055\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":760}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Red, 60\",\"ProductID\":760,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50R-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.90,\"rowguid\":\"664867e5-4ab3-4783-96f9-42efde92f49b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":761}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Red, 62\",\"ProductID\":761,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50R-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.00,\"rowguid\":\"1da14e09-6d71-4e2a-9ee9-1bdfdfd8a109\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":762}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Red, 44\",\"ProductID\":762,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":18.77,\"rowguid\":\"f247aaae-12e3-4048-a37b-cce4a8999e81\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":763}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Red, 48\",\"ProductID\":763,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.13,\"rowguid\":\"3da5fa6e-9e0f-4896-ac10-948c27f4eb79\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":764}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Red, 52\",\"ProductID\":764,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.42,\"rowguid\":\"07cfe1ea-8a37-4d2a-835f-bc8d37e564af\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":765}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Black, 58\",\"ProductID\":765,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50B-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.79,\"rowguid\":\"74645b12-3ce9-49a6-bd96-2cd814b37420\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":766}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Black, 60\",\"ProductID\":766,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50B-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.90,\"rowguid\":\"a2db196d-6640-49ea-a84f-2e87ca6f50c6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":767}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Black, 62\",\"ProductID\":767,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50B-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.00,\"rowguid\":\"c82a8309-63d3-4c0c-ad71-e91272397095\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":768}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Black, 44\",\"ProductID\":768,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":18.77,\"rowguid\":\"11d563ac-115c-4f0d-a1e5-e946eee8b38b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":769}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Black, 48\",\"ProductID\":769,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.13,\"rowguid\":\"a7e2179e-137c-497e-a5e6-c9ea64935fb0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":770}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":782.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-650 Black, 52\",\"ProductID\":770,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":30}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R50B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":486.7066,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.42,\"rowguid\":\"136e2865-e0da-4624-963a-31349279ab1a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":771}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3399.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Silver, 38\",\"ProductID\":771,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82S-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":1912.1544,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.35,\"rowguid\":\"ca74b54e-fc30-4464-8b83-019bfd1b2dbb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":772}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3399.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Silver, 42\",\"ProductID\":772,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":1912.1544,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.77,\"rowguid\":\"bbfff5a5-4bdc-49a9-a5ad-7584adffe808\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":773}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3399.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Silver, 44\",\"ProductID\":773,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82S-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":1912.1544,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":21.13,\"rowguid\":\"155fd77e-d6d6-43f0-8a5b-4a8305eb45cd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":774}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3399.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Silver, 48\",\"ProductID\":774,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82S-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":1912.1544,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":21.42,\"rowguid\":\"ba5551df-c9ee-4b43-b3ca-8c19d0f9384d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":775}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3374.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Black, 38\",\"ProductID\":775,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82B-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":1898.0944,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.35,\"rowguid\":\"dea33347-fcd3-4346-aa0f-068cd6b3ad06\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":776}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3374.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Black, 42\",\"ProductID\":776,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82B-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":1898.0944,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.77,\"rowguid\":\"02935111-a546-4c6d-941f-be12d42c158e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":777}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3374.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Black, 44\",\"ProductID\":777,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":1898.0944,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":21.13,\"rowguid\":\"7920bc3b-8fd4-4610-93d2-e693a66b6474\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":778}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3374.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-100 Black, 48\",\"ProductID\":778,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":19}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M82B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1025420400000-0700)\\/\",\"SellStartDate\":\"\\/Date(993970800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":1898.0944,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":21.42,\"rowguid\":\"1b486300-7e64-4c5d-a9ba-a8368e20c5a0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":779}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2319.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-200 Silver, 38\",\"ProductID\":779,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":20}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M68S-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":1265.6195,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":23.35,\"rowguid\":\"3a45d835-abae-4806-ac03-c53abcd3b974\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":780}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2319.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-200 Silver, 42\",\"ProductID\":780,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":20}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M68S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":1265.6195,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":23.77,\"rowguid\":\"ce4849b4-56e6-4b50-808b-9bde67cc4704\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":781}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2319.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-200 Silver, 46\",\"ProductID\":781,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":20}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M68S-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":1265.6195,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":24.13,\"rowguid\":\"20799030-420c-496a-9922-09189c2b457e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":782}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2294.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-200 Black, 38\",\"ProductID\":782,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":20}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M68B-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":1251.9813,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":23.35,\"rowguid\":\"82cb8f9b-b8bb-4841-98d3-bcdb807c4dd8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":783}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2294.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-200 Black, 42\",\"ProductID\":783,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":20}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M68B-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":1251.9813,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":23.77,\"rowguid\":\"2b0af5b9-7571-4621-b760-47df599f9650\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":784}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2294.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-200 Black, 46\",\"ProductID\":784,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":20}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M68B-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":1251.9813,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":24.13,\"rowguid\":\"33f42284-e216-4b98-ba48-b4ee18a01fa0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":785}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1079.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-300 Black, 38\",\"ProductID\":785,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":21}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M47B-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":598.4354,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.35,\"rowguid\":\"f06c2cbf-0901-4c08-80ed-fb4e87171b3b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":786}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1079.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-300 Black, 40\",\"ProductID\":786,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":21}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M47B-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":598.4354,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.77,\"rowguid\":\"580d4322-be03-4c91-83d2-ee33ec6008ab\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":787}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1079.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-300 Black, 44\",\"ProductID\":787,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":21}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M47B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":598.4354,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":26.13,\"rowguid\":\"07c2a548-0452-47b4-9dce-6edb0a30c85e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":788}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1079.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-300 Black, 48\",\"ProductID\":788,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":21}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M47B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":598.4354,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":26.42,\"rowguid\":\"16078dbe-388d-4c18-aa8f-0c8f48035468\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":789}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Red, 44\",\"ProductID\":789,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89R-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":1518.7864,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":14.77,\"rowguid\":\"0aa71ad6-afaf-43c6-9745-35d815b50a5b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":790}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Red, 48\",\"ProductID\":790,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89R-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":1518.7864,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.13,\"rowguid\":\"115ddade-70e3-43f9-80dc-638daea271c4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":791}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Red, 52\",\"ProductID\":791,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89R-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":1518.7864,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.42,\"rowguid\":\"c9fd1df4-9512-420a-b379-067108033b75\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":792}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Red\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Red, 58\",\"ProductID\":792,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89R-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":1554.9479,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.79,\"rowguid\":\"df629c11-8d8b-4774-9d52-ecb64dc95212\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":793}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Black, 44\",\"ProductID\":793,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":1554.9479,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":14.77,\"rowguid\":\"1ff419b5-52af-4f7e-aeae-4fec5e99de35\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":794}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Black, 48\",\"ProductID\":794,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":1554.9479,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.13,\"rowguid\":\"9d165ddf-8f5d-41c7-9bb8-13f41a3d1f62\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":795}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Black, 52\",\"ProductID\":795,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":1554.9479,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.42,\"rowguid\":\"74fe3957-cbc3-450a-ab92-849bd13530e2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":796}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2443.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-250 Black, 58\",\"ProductID\":796,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":26}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R89B-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":1554.9479,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.68,\"rowguid\":\"1c530fe8-a169-4adc-b3dc-b11a48245e63\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":797}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1120.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-550-W Yellow, 38\",\"ProductID\":797,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":29}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R64Y-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":713.0798,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":17.35,\"rowguid\":\"aad81532-a572-49a5-83c3-dfa9e3b4fea6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":798}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1120.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-550-W Yellow, 40\",\"ProductID\":798,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":29}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R64Y-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":713.0798,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":17.77,\"rowguid\":\"a35a1c35-c128-4697-951e-4199062e78f3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":799}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1120.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-550-W Yellow, 42\",\"ProductID\":799,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":29}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R64Y-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":713.0798,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":18.13,\"rowguid\":\"a359ab09-16f2-4726-a60f-0dfdb1c9527e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":800}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1120.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-550-W Yellow, 44\",\"ProductID\":800,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":29}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R64Y-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":713.0798,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":18.42,\"rowguid\":\"0a7028fb-ff06-4d38-aaa1-b64816278165\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":801}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1120.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-550-W Yellow, 48\",\"ProductID\":801,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":29}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R64Y-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":713.0798,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":18.68,\"rowguid\":\"c90cc877-804c-4ce7-afc3-4c8791a13dfb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":802}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":148.2200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Fork\",\"ProductID\":802,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":104}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FK-1639\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":10}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":65.8097,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fb8502be-07eb-4134-ab06-c3a9959a52ae\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":803}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":175.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Fork\",\"ProductID\":803,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":105}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FK-5136\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":10}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":77.9176,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"f5fa4e2f-b976-48a4-bf79-85632f697d2e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":804}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":229.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Fork\",\"ProductID\":804,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":106}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FK-9939\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":10}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":101.8936,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"553229b3-1ad9-4a71-a21c-2af4332cfce9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":805}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":34.2000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Headset\",\"ProductID\":805,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":59}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HS-0296\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":11}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":15.1848,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"bb6bd7b3-a34d-4d64-822e-781fa6838e19\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":806}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":102.2900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Headset\",\"ProductID\":806,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":60}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HS-2451\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":11}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":45.4168,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"23b5d52b-8c29-4059-b899-75c53b5ee2e6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":807}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":124.7300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Headset\",\"ProductID\":807,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":61}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HS-3479\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":11}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":55.3801,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"12e4d5e8-79ed-4bcb-a532-6275d1a93417\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":808}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":44.5400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Handlebars\",\"ProductID\":808,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":52}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-M243\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":19.7758,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b59b7bf2-7afc-4a74-b063-f942f1e0da19\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":809}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":61.9200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Handlebars\",\"ProductID\":809,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":54}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-M763\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":27.4925,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ae6020df-d9c9-4d34-9795-1f80e6bbf5a5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":810}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":120.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Handlebars\",\"ProductID\":810,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":55}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-M918\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":53.3999,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6aa0f921-0f09-4f99-8d3c-335946873553\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":811}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":44.5400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Handlebars\",\"ProductID\":811,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":56}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-R504\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":19.7758,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"acdae407-b40e-435e-8e84-1bfc33013e81\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":812}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":61.9200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Handlebars\",\"ProductID\":812,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":57}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-R720\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":27.4925,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"33fa1a03-38d6-405e-be9b-eea92f3d204f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":813}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":120.2700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Handlebars\",\"ProductID\":813,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":58}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-R956\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":53.3999,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5c5327b9-ad2e-4c33-8d74-edf49e0c2dd8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":814}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":348.7600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame - Black, 38\",\"ProductID\":814,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":15}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63B-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":185.8193,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.73,\"rowguid\":\"0d879312-a7d3-441d-9d23-b6550bab3814\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":815}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":60.7450,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Front Wheel\",\"ProductID\":815,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":42}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-M423\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":26.9708,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d7b1d161-182e-44f6-a9ff-9c1eba76613b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":816}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":209.0250,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Front Wheel\",\"ProductID\":816,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":45}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-M762\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":92.8071,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5e3e5033-9a77-4dca-8b7f-dfed78efa08a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":817}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":300.2150,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Front Wheel\",\"ProductID\":817,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":46}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-M928\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":133.2955,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3c8ffff6-a8dc-45a3-963b-a6284ced7596\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":818}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":85.5650,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Front Wheel\",\"ProductID\":818,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":49}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-R623\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":37.9909,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":900.00,\"rowguid\":\"9e66de78-decb-438a-b9a9-023c773c60a2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":819}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":248.3850,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Front Wheel\",\"ProductID\":819,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":50}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-R762\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":110.2829,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":850.00,\"rowguid\":\"6ea94fbf-b9aa-43fc-84e8-91d508dde751\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":820}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":330.0600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Front Wheel\",\"ProductID\":820,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":51}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-R820\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":146.5466,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":650.00,\"rowguid\":\"727a3cd5-8d40-4884-a7e4-dfd3ffdeb164\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":821}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":218.0100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring Front Wheel\",\"ProductID\":821,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":44}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FW-T905\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":96.7964,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"8d1cdb07-4fa1-4ac1-840f-a16c3dca5009\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":822}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame-W - Yellow, 38\",\"ProductID\":822,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":17}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72Y-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":360.9428,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.18,\"rowguid\":\"22976fa7-0ad0-40f9-b4f9-ba10279ea1a3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":823}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":87.7450,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Rear Wheel\",\"ProductID\":823,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":123}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-M423\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":38.9588,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e6c39f58-995f-4786-a309-df3812d8b30f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":824}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":236.0250,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Rear Wheel\",\"ProductID\":824,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":124}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-M762\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":104.7951,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b2f7cf9b-a7bf-49ab-9cca-9f6e791cd693\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":825}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":327.2150,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Rear Wheel\",\"ProductID\":825,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":125}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-M928\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":145.2835,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"35d02edc-1120-41fb-b5df-8655a93b3353\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":826}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":112.5650,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Rear Wheel\",\"ProductID\":826,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":126}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-R623\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":49.9789,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":1050.00,\"rowguid\":\"78d01117-8dcd-465f-9dc7-94a2cdc35582\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":827}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":275.3850,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Rear Wheel\",\"ProductID\":827,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":77}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-R762\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":122.2709,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":1000.00,\"rowguid\":\"cf739f9a-0af0-4810-b229-c431a31d7779\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":828}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":357.0600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Rear Wheel\",\"ProductID\":828,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":78}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-R820\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":158.5346,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":890.00,\"rowguid\":\"8f262a37-43aa-4ad5-ae1f-8bd6967d8e9b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":829}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":245.0100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring Rear Wheel\",\"ProductID\":829,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":43}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RW-T905\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":17}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":108.7844,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"30d2254d-e26d-4586-b4c5-e8ccb8a059b8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":830}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":348.7600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame - Black, 40\",\"ProductID\":830,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":14}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63B-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":185.8193,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.77,\"rowguid\":\"aed1c9c4-c139-45d2-b38e-0a0a9f518665\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":831}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":348.7600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame - Black, 44\",\"ProductID\":831,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":14}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":185.8193,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.81,\"rowguid\":\"3f2135d4-ec5f-4e30-bde8-5444518f0637\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":832}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":348.7600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame - Black, 48\",\"ProductID\":832,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":14}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":185.8193,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.85,\"rowguid\":\"c2009b69-f15a-47ec-b710-1809d299318a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":833}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame-W - Yellow, 40\",\"ProductID\":833,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":17}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72Y-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":360.9428,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.22,\"rowguid\":\"22df26f2-60bc-493e-a14a-5500633e9f7e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":834}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame-W - Yellow, 42\",\"ProductID\":834,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":17}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72Y-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":360.9428,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.26,\"rowguid\":\"207b54da-5404-415d-8578-9a45082e3bf1\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":835}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame-W - Yellow, 44\",\"ProductID\":835,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":17}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72Y-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":360.9428,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.30,\"rowguid\":\"a0fad492-ac24-4fcf-8d2a-d21d06386ae1\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":836}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":594.8300,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Frame-W - Yellow, 48\",\"ProductID\":836,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":17}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R72Y-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":360.9428,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.34,\"rowguid\":\"8487bfe0-2138-471e-9c6d-fdb3a67e7d86\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":837}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Black, 62\",\"ProductID\":837,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92B-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.30,\"rowguid\":\"5dce9c8c-fb94-46f8-b826-11658f6a3682\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":838}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Black, 44\",\"ProductID\":838,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.12,\"rowguid\":\"e4902656-a4bc-4b08-9d47-4f3da0fd76c3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":839}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Black, 48\",\"ProductID\":839,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.16,\"rowguid\":\"557b603b-407b-41a4-ae34-4f7747866dba\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":840}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1431.5000,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Frame - Black, 52\",\"ProductID\":840,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":6}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-R92B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":14}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":868.6342,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.20,\"rowguid\":\"0ed082b3-fbba-43af-9149-8741b9fc78c8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":841}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":59.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Sports Shorts, S\",\"ProductID\":841,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":13}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-M897-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":24.7459,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"34b08c1f-99d1-43c4-8ef7-2cd754b6665d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":842}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Grey\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":125.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-Panniers, Large\",\"ProductID\":842,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":120}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PA-T100\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":35}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":51.5625,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"56334fff-91d4-495e-bf98-933bc1010f23\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":843}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":25.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Cable Lock\",\"ProductID\":843,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":115}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LO-C100\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":34}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":10.3125,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"56ffd7b9-1014-4640-b1bd-b2649589b4d7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":844}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":19.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Minipump\",\"ProductID\":844,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":116}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PU-0452\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":36}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":8.2459,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"aaf7a076-9ee3-46bf-a69f-414d847e858b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":845}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":24.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain Pump\",\"ProductID\":845,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":117}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PU-M044\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":36}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":10.3084,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"57169f80-fafb-4678-8f51-fe1f131d0c83\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":846}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":13.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Taillights - Battery-Powered\",\"ProductID\":846,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":108}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LT-T990\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":33}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":5.7709,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3c617b87-50a5-434c-a0d3-22314b7027ee\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":847}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":34.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Headlights - Dual-Beam\",\"ProductID\":847,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":109}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LT-H902\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":33}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":14.4334,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"417db6cb-f38f-4b0d-87e7-e1ebf7456c3a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":848}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":44.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Headlights - Weatherproof\",\"ProductID\":848,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":110}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"LT-H903\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":33}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":null,\"StandardCost\":18.5584,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fc362c1a-4b9c-4d5f-a6d3-0775846c61f0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":849}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":59.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Sports Shorts, M\",\"ProductID\":849,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":13}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-M897-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":24.7459,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"db37b435-74b9-43d3-b363-abbead107bc4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":850}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":59.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Sports Shorts, L\",\"ProductID\":850,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":13}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-M897-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":24.7459,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"714184c5-669b-4cf1-a802-30e7b1ea7722\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":851}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":59.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Sports Shorts, XL\",\"ProductID\":851,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":13}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-M897-X\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"XL\",\"StandardCost\":24.7459,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"bd3ffe40-fe2e-44cb-a4e0-81786c3a751f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":852}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":74.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Women's Tights, S\",\"ProductID\":852,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":38}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TG-W091-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":24}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":30.9334,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7de86c98-4f5b-4155-8572-c977f14ebaeb\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":853}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":74.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Women's Tights, M\",\"ProductID\":853,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":38}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TG-W091-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":24}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":30.9334,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"4d8e186c-b8c9-4c64-b411-4995dd87e316\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":854}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":74.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Women's Tights, L\",\"ProductID\":854,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":38}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TG-W091-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":24}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":30.9334,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e378c2f3-54f6-4ea9-b049-e8bb32b5bdfd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":855}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":89.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Bib-Shorts, S\",\"ProductID\":855,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":12}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SB-M891-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":18}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":37.1209,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9f60af1e-4c11-4e90-baea-48e834e8b4c2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":856}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":89.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Bib-Shorts, M\",\"ProductID\":856,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":12}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SB-M891-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":18}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":37.1209,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e0cbec04-f4fc-450f-9780-f8ea7691febd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":857}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Multi\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":89.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Men's Bib-Shorts, L\",\"ProductID\":857,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":12}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SB-M891-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":18}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":37.1209,\"Style\":\"M \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e1df75a4-9986-489e-a5de-ad3da824eb5e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":858}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":24.4900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Half-Finger Gloves, S\",\"ProductID\":858,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":4}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"GL-H102-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":20}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":9.1593,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9e1db5c3-539d-4061-9433-d762dc195cd8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":859}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":24.4900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Half-Finger Gloves, M\",\"ProductID\":859,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":4}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"GL-H102-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":20}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":9.1593,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9d458fd5-392d-4ab1-afef-6a5548e48858\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":860}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":24.4900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Half-Finger Gloves, L\",\"ProductID\":860,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":4}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"GL-H102-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":20}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":9.1593,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"fa710215-925f-4959-81df-538e72a6a255\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":861}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":37.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Full-Finger Gloves, S\",\"ProductID\":861,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":3}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"GL-F110-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":20}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":15.6709,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"76fac097-1fb3-456b-8fb9-2c7a613771b4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":862}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":37.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Full-Finger Gloves, M\",\"ProductID\":862,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":3}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"GL-F110-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":20}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":15.6709,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1084221e-1890-443e-9d87-afcad6358355\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":863}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":37.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Full-Finger Gloves, L\",\"ProductID\":863,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":3}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"GL-F110-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":20}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":\"\\/Date(1056956400000-0700)\\/\",\"SellStartDate\":\"\\/Date(1025506800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":15.6709,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6116f9d4-8a1d-4022-a642-9c445c197203\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":864}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Blue\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":63.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Classic Vest, S\",\"ProductID\":864,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":1}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"VE-C304-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":25}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":23.7490,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"eb423ef3-409d-46fe-b35b-d69970820314\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":865}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Blue\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":63.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Classic Vest, M\",\"ProductID\":865,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":1}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"VE-C304-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":25}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":23.7490,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2e52f96e-64a1-4069-911c-e3fd6e094a1e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":866}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Blue\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":63.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Classic Vest, L\",\"ProductID\":866,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":1}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"VE-C304-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":25}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":23.7490,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3211f5a8-b6c4-48bd-9aa4-d69cb40d97dd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":867}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":69.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Women's Mountain Shorts, S\",\"ProductID\":867,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":37}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-W890-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":26.1763,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"22616fd2-b99f-4f7d-acf6-33dff66d42d2\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":868}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":69.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Women's Mountain Shorts, M\",\"ProductID\":868,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":37}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-W890-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":26.1763,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"968e3610-e583-42e8-8ab6-484a799b1774\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":869}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Black\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":69.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Women's Mountain Shorts, L\",\"ProductID\":869,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":37}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SH-W890-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":22}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":26.1763,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"1a66b244-5ca0-4153-b539-ae048d14faec\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":870}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":4.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Water Bottle - 30 oz.\",\"ProductID\":870,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":111}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"WB-H098\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":28}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":1.8663,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"834e8d1a-69a7-4c42-8b68-fa08d9ec9e5b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":871}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":9.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain Bottle Cage\",\"ProductID\":871,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":112}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BC-M005\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":28}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":3.7363,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"684491de-63f8-4632-90e3-36773c4e63bd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":872}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":8.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road Bottle Cage\",\"ProductID\":872,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":113}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BC-R205\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":28}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":3.3623,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"601b1fe8-d4d0-4cfb-9379-29481cc05291\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":873}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2.2900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Patch Kit\\/8 Patches\",\"ProductID\":873,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":114}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PK-7098\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":0.8565,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"36e638e4-68df-411b-930a-daad57221aa6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":874}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"White\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":8.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Racing Socks, M\",\"ProductID\":874,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":24}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SO-R809-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":23}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":3.3623,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"b9c7eb0a-1dd1-4a1d-b4c3-1dad83a8ea7e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":875}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"White\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":8.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Racing Socks, L\",\"ProductID\":875,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":24}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SO-R809-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":23}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":3.3623,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"c0a16305-74b7-4fae-b5c9-3e8bd0e44762\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":876}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":120.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hitch Rack - 4-Bike\",\"ProductID\":876,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":118}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RA-H123\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":26}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":44.8800,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7a0c4bbd-9679-4f59-9ebc-9daf3439a38a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":877}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":7.9500,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Bike Wash - Dissolver\",\"ProductID\":877,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":119}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"CL-9009\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":29}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":2.9733,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"3c40b5ad-e328-4715-88a7-ec3220f02acf\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":878}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":21.9800,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Fender Set - Mountain\",\"ProductID\":878,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":121}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FE-6654\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":30}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":8.2205,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"e6e76c7f-c145-4cad-a9e8-b1e4e845a2c0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":879}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":159.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"All-Purpose Bike Stand\",\"ProductID\":879,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":122}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"ST-1401\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":27}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":59.4660,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"c7bb564b-a637-40f5-b21b-cbf7e4f713be\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":880}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":54.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Hydration Pack - 70 oz.\",\"ProductID\":880,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":107}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HY-1023-70\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":32}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"70\",\"StandardCost\":20.5663,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a99d90c0-05e2-4e44-ad90-c55c9f0784de\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":881}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Yellow\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":53.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Short-Sleeve Classic Jersey, S\",\"ProductID\":881,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":32}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SJ-0194-S\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"S\",\"StandardCost\":41.5723,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"05b2e20f-2399-4cb3-9b49-b28d6649c104\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":882}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Yellow\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":53.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Short-Sleeve Classic Jersey, M\",\"ProductID\":882,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":32}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SJ-0194-M\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"M\",\"StandardCost\":41.5723,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"bbbf003b-367d-4d71-af71-10f50b6234a0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":883}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Yellow\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":53.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Short-Sleeve Classic Jersey, L\",\"ProductID\":883,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":32}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SJ-0194-L\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"L\",\"StandardCost\":41.5723,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2d9f59b8-9f24-46eb-98ad-553e48bb9db9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":884}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Yellow\",\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":53.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Short-Sleeve Classic Jersey, XL\",\"ProductID\":884,\"ProductLine\":\"S \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":32}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SJ-0194-X\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":21}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":3,\"SafetyStockLevel\":4,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"XL\",\"StandardCost\":41.5723,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"906d42f6-c21f-4d20-b528-02ffdb55fd1e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":885}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Yellow, 60\",\"ProductID\":885,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98Y-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.08,\"rowguid\":\"c49679bd-96a9-4176-a7ed-5bc6d6444647\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":886}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Yellow, 62\",\"ProductID\":886,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67Y-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.20,\"rowguid\":\"8d4d52a6-8abc-4c05-be4d-c067faf1a64e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":887}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Yellow, 46\",\"ProductID\":887,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98Y-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.96,\"rowguid\":\"137675a7-34cd-4b7a-abe1-4e0eeb79b65d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":888}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Yellow, 50\",\"ProductID\":888,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98Y-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.00,\"rowguid\":\"105ec6e5-30c5-4fe3-a08b-cb324c85323d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":889}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Yellow, 54\",\"ProductID\":889,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98Y-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.04,\"rowguid\":\"12b1f317-c39b-48d0-b1a7-8018c60aea53\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":890}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Blue, 46\",\"ProductID\":890,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98U-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.96,\"rowguid\":\"8bbd3437-a58b-41a0-9503-fc14b23e7678\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":891}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Blue, 50\",\"ProductID\":891,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98U-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.00,\"rowguid\":\"c4244f0c-abce-451b-a895-83c0e6d1f448\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":892}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Blue, 54\",\"ProductID\":892,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98U-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.04,\"rowguid\":\"e9aae947-6bc3-4909-8937-3e1cdcec8a8f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":893}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1003.9100,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Frame - Blue, 60\",\"ProductID\":893,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":7}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T98U-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":601.7437,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.08,\"rowguid\":\"b01951a4-a581-4d10-9dc2-515da180f1b8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":894}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":121.4600,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Rear Derailleur\",\"ProductID\":894,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":127}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RD-2308\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":9}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":53.9282,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":215.00,\"rowguid\":\"5ebfcf02-4e3e-443a-ad60-5aeef28dac76\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":895}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Blue, 50\",\"ProductID\":895,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67U-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.10,\"rowguid\":\"b78eccca-fa88-4071-9110-410585127e46\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":896}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Blue, 54\",\"ProductID\":896,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67U-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.14,\"rowguid\":\"0ff799c9-dd11-4b81-aaf7-65410017405b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":897}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Blue, 58\",\"ProductID\":897,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67U-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.16,\"rowguid\":\"ccd4fa47-7194-4bd0-909b-1fa4bd7916a7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":898}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Blue, 62\",\"ProductID\":898,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67U-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.20,\"rowguid\":\"08a211a5-dcd2-42d0-9276-64d92d4890a6\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":899}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Yellow, 44\",\"ProductID\":899,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67Y-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.02,\"rowguid\":\"109cb7bc-6ec6-4a36-85c8-55b843b2ab12\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":900}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Yellow, 50\",\"ProductID\":900,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67Y-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.10,\"rowguid\":\"285fd682-c647-49d1-b8f3-368a43c9cda0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":901}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Yellow, 54\",\"ProductID\":901,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67Y-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.14,\"rowguid\":\"2536e3b2-d4da-49e6-965a-f612c2c8914f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":902}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Yellow, 58\",\"ProductID\":902,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67Y-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.16,\"rowguid\":\"5d17ff1c-f50e-438f-a4e9-7c400fb762e7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":903}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":333.4200,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Frame - Blue, 44\",\"ProductID\":903,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":10}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-T67U-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":16}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":199.8519,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.02,\"rowguid\":\"e9c17be7-f4dc-465e-ab73-c0198f37bfdd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":904}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":364.0900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame-W - Silver, 40\",\"ProductID\":904,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":15}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63S-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":199.3757,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.77,\"rowguid\":\"a7dde43e-f7d5-4075-a0c1-c866ad7aa154\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":905}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":364.0900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame-W - Silver, 42\",\"ProductID\":905,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":15}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":199.3757,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.81,\"rowguid\":\"d4a2fcad-1e63-4ebd-863c-5a7c48d5b8d9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":906}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":364.0900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame-W - Silver, 46\",\"ProductID\":906,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":15}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63S-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":199.3757,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.85,\"rowguid\":\"82025c63-7b28-412d-92c1-408e0e6ae646\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":907}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":106.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Rear Brakes\",\"ProductID\":907,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":128}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"RB-9231\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":6}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":47.2860,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":317.00,\"rowguid\":\"5946f163-93f0-4141-b17e-55d9778cc274\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":908}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":27.1200,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Seat\\/Saddle\",\"ProductID\":908,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":79}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-M236\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":12.0413,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"4dab53c5-31e7-47d6-a5a0-940f8d4dad22\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":909}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":39.1400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Seat\\/Saddle\",\"ProductID\":909,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":80}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-M798\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":17.3782,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"30222244-0fd8-4d28-8448-f2e658bf52bd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":910}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":52.6400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Seat\\/Saddle\",\"ProductID\":910,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":81}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-M940\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":23.3722,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a96a5024-87de-488a-bf81-bc0c81f6cd18\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":911}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":27.1200,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Seat\\/Saddle\",\"ProductID\":911,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":82}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-R581\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":12.0413,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"31b9bc62-792b-4e7a-a24d-411dc76e0271\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":912}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":39.1400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Seat\\/Saddle\",\"ProductID\":912,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":83}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-R908\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":17.3782,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"49845afe-a8cc-4354-a5d4-09d35bf7fb9e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":913}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":52.6400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Seat\\/Saddle\",\"ProductID\":913,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":84}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-R995\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":23.3722,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"2befe629-4b2a-41a1-a009-df13ead69105\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":914}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":27.1200,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Seat\\/Saddle\",\"ProductID\":914,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":66}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-T312\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":12.0413,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"7874a1d6-7a5b-412f-a2eb-c2f457b9603d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":915}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":39.1400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Touring Seat\\/Saddle\",\"ProductID\":915,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":65}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-T762\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":17.3782,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"072acb72-7796-4bd0-9bbb-6efc29ac336c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":916}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":52.6400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Seat\\/Saddle\",\"ProductID\":916,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":67}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"SE-T924\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":15}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":23.3722,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"0e158724-934d-4a64-991f-94fec00bdea8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":917}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":264.0500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Silver, 42\",\"ProductID\":917,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":144.5938,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.92,\"rowguid\":\"37bd4d2b-346b-4c6b-8f3f-85c084282529\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":918}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":264.0500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Silver, 44\",\"ProductID\":918,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21S-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":144.5938,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.96,\"rowguid\":\"393a4d00-7428-41f0-a48a-af26b00e9a9c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":919}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":264.0500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Silver, 48\",\"ProductID\":919,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21S-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":144.5938,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.00,\"rowguid\":\"8dfef6f2-91a8-4884-8949-b2551218b37a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":920}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":264.0500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Silver, 52\",\"ProductID\":920,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21S-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":144.5938,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.04,\"rowguid\":\"f230baac-5951-4eb1-b1e8-94c2ca2b37fa\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":921}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":4.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain Tire Tube\",\"ProductID\":921,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":92}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TT-M928\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":1.8663,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"01a8c3fc-ed52-458e-a634-d5b6e2accfed\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":922}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":3.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road Tire Tube\",\"ProductID\":922,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":93}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TT-R982\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":1.4923,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ea442bd7-f69b-42d9-aa71-95e10b648f52\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":923}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":4.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring Tire Tube\",\"ProductID\":923,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":94}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TT-T092\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":1.8663,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"179fec38-cab5-4a47-bcff-31cfc9e43a3c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":924}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":249.7900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Black, 42\",\"ProductID\":924,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21B-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":136.7850,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.92,\"rowguid\":\"faabd7fb-cb35-4bad-8857-ec71468686ad\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":925}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":249.7900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Black, 44\",\"ProductID\":925,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":136.7850,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.96,\"rowguid\":\"47ab0300-7b55-4d35-a786-80190976b9b5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":926}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":249.7900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Black, 48\",\"ProductID\":926,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":136.7850,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.00,\"rowguid\":\"408435aa-15c0-41e5-981f-32a8226af15f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":927}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":249.7900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Black, 52\",\"ProductID\":927,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":136.7850,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":3.04,\"rowguid\":\"4800f4e6-99ea-4afd-a392-2cb05265d0d4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":928}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":24.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Tire\",\"ProductID\":928,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":85}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-M267\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":9.3463,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"76060a93-949c-48ea-9b31-a593d6c14983\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":929}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":29.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Tire\",\"ProductID\":929,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":86}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-M602\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":11.2163,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"daff9e11-6254-432d-8c4f-f06e52687184\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":930}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":35.0000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Tire\",\"ProductID\":930,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":87}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-M823\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":13.0900,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"ddad25f5-0445-4b5c-8466-6446930ad8b8\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":931}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":21.4900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Tire\",\"ProductID\":931,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":88}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-R092\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":8.0373,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"15b569a6-d172-42c2-a420-62ab5946cc80\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":932}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":24.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Tire\",\"ProductID\":932,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":89}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-R628\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":9.3463,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"d1016cc5-c12b-4f05-915c-70fa062e4a64\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":933}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":32.6000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Road Tire\",\"ProductID\":933,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":90}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-R982\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":12.1924,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"c86de56a-5048-4b89-b7c7-56bc75c9bcee\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":934}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":null,\"DaysToManufacture\":0,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":28.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring Tire\",\"ProductID\":934,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":91}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"TI-T723\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":37}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":10.8423,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"9d5ca300-5302-41e1-bca5-8ce5b93f26a5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":935}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":40.4900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Pedal\",\"ProductID\":935,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":62}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-M282\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":17.9776,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":218.00,\"rowguid\":\"9fdd0c65-b2b0-4c6c-8704-3a9747be5174\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":936}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":62.0900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Pedal\",\"ProductID\":936,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":63}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-M340\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":27.5680,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":215.00,\"rowguid\":\"274c86dc-439e-4469-9de8-7e9bd6455d0d\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":937}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":80.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Mountain Pedal\",\"ProductID\":937,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":64}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-M562\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":35.9596,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":185.00,\"rowguid\":\"a05464e8-6b4d-42b3-a4d6-8683136f4b66\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":938}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":40.4900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Road Pedal\",\"ProductID\":938,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":68}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-R347\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":17.9776,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":189.00,\"rowguid\":\"2c7dd8c3-4c55-475f-ba58-f4baca520d72\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":939}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":62.0900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Road Pedal\",\"ProductID\":939,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":69}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-R563\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":27.5680,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":168.00,\"rowguid\":\"216ad46f-bc3f-4862-b0e9-2e261e5a6059\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":940}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":80.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028235510-0800)\\/\",\"Name\":\"HL Road Pedal\",\"ProductID\":940,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":70}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-R853\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":35.9596,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":149.00,\"rowguid\":\"44e96967-ab99-41ed-8b41-5bc70a5ca1a9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":941}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\\/Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":80.9900,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring Pedal\",\"ProductID\":941,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":53}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"PD-T852\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":13}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":35.9596,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"6967c816-3ebb-45fa-8547-cef00e08573e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":942}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":364.0900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Mountain Frame-W - Silver, 38\",\"ProductID\":942,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":15}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M63S-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":199.3757,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.73,\"rowguid\":\"ba3646b0-1487-426e-ab4e-57d42e6f9233\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":943}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":249.7900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Black, 40\",\"ProductID\":943,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21B-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":136.7850,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.88,\"rowguid\":\"3050e4df-2bba-4c2b-bdcc-b4c89972db1c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":944}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":2,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":264.0500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Mountain Frame - Silver, 40\",\"ProductID\":944,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":8}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FR-M21S-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":12}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":144.5938,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":2.88,\"rowguid\":\"756f862e-40cc-4dfc-b189-716d0dda5ff9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":945}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":91.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Front Derailleur\",\"ProductID\":945,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":103}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FD-2342\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":9}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":40.6216,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":88.00,\"rowguid\":\"448e9e7b-9548-4a4c-abb3-853686aa7517\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":946}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":46.0900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Touring Handlebars\",\"ProductID\":946,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":47}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-T721\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":20.4640,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"a2f1352e-45d0-42c4-aef3-60033073bb66\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":947}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":91.5700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Touring Handlebars\",\"ProductID\":947,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":48}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"HB-T928\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":4}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":40.6571,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"cb524e92-4fa8-4f6c-9993-60796856c654\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":948}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":106.5000,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Front Brakes\",\"ProductID\":948,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":102}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"FB-9873\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":6}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":47.2860,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":317.00,\"rowguid\":\"c1813164-1b4b-42d1-9007-4e5f9aee0e19\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":949}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":175.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Crankset\",\"ProductID\":949,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":99}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"CS-4759\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":8}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":77.9176,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":600.00,\"rowguid\":\"41d47371-4374-46d3-8d61-b07616ec54f0\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":950}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":256.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Crankset\",\"ProductID\":950,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":100}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"CS-6583\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":8}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":113.8816,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":635.00,\"rowguid\":\"d3a7a02c-a3d5-4a04-9454-0c4e43772b78\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":951}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Black\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":404.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Crankset\",\"ProductID\":951,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":101}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"CS-9183\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":8}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":179.8156,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":575.00,\"rowguid\":\"2c4a8956-7b72-48fe-b028-699e117b1daa\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":952}],\"EntitySetName\":\"Products\"},\"Class\":null,\"Color\":\"Silver\",\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":20.2400,\"MakeFlag\":false,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Chain\",\"ProductID\":952,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":98}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"CH-0234\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":7}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":8.9866,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":null},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":null,\"rowguid\":\"5d27e2a5-27ec-4ccb-ba2c-fc980ffe6708\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":953}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1214.8500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-2000 Blue, 60\",\"ProductID\":953,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":35}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T44U-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":755.1508,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.90,\"rowguid\":\"f1bb3957-8d27-47f3-91ec-c71822d11436\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":954}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Yellow, 46\",\"ProductID\":954,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79Y-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.13,\"rowguid\":\"83b77413-8c8a-4af1-93e4-136edb7ff15f\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":955}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Yellow, 50\",\"ProductID\":955,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79Y-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.42,\"rowguid\":\"5ec991ad-8761-4a61-a318-312df3a78604\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":956}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Yellow, 54\",\"ProductID\":956,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79Y-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.68,\"rowguid\":\"1220b0f0-064d-46b7-8507-1fa758b77b9c\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":957}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Yellow, 60\",\"ProductID\":957,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79Y-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.90,\"rowguid\":\"bcd1a5a9-6140-4dc9-9620-689dc7e4c155\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":958}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Blue, 54\",\"ProductID\":958,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18U-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":29.68,\"rowguid\":\"a3ee6897-52fe-42e4-92ec-7a91e7bb905a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":959}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Blue, 58\",\"ProductID\":959,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18U-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":29.90,\"rowguid\":\"23d89cee-9f44-4f3e-b289-63de6ba2b737\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":960}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Blue, 62\",\"ProductID\":960,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18U-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":30.00,\"rowguid\":\"060192c9-bcd9-4260-b729-d6bcfadfb08e\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":961}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Yellow, 44\",\"ProductID\":961,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18Y-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.77,\"rowguid\":\"5646c15a-68ad-4234-b328-254706cbccc5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":962}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Yellow, 50\",\"ProductID\":962,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18Y-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":29.13,\"rowguid\":\"df85e805-af87-4fab-a668-c80f2a5b8a69\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":963}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Yellow, 54\",\"ProductID\":963,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18Y-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":29.42,\"rowguid\":\"192becd1-f465-4194-88a2-ee57fed3a3c5\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":964}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Yellow, 58\",\"ProductID\":964,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18Y-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":29.79,\"rowguid\":\"bed79f64-a53d-44a3-ace8-2baa425a5a54\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":965}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Yellow, 62\",\"ProductID\":965,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18Y-62\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"62\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":30.00,\"rowguid\":\"d28b3872-5173-40a4-b12f-655524386cc7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":966}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Blue, 46\",\"ProductID\":966,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79U-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.13,\"rowguid\":\"86990d54-6efb-4c53-9974-6c3b0297e222\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":967}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Blue, 50\",\"ProductID\":967,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79U-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.42,\"rowguid\":\"68c0a818-2985-46eb-8255-0fb70919fa24\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":968}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Blue, 54\",\"ProductID\":968,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79U-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.68,\"rowguid\":\"12280a8c-7578-4367-ba71-214bcc1e4792\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":969}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":2384.0700,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-1000 Blue, 60\",\"ProductID\":969,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":34}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T79U-60\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"60\",\"StandardCost\":1481.9379,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":25.90,\"rowguid\":\"dd70cf36-449a-43fd-839d-a84ee14c849a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":970}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1214.8500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-2000 Blue, 46\",\"ProductID\":970,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":35}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T44U-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":755.1508,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.13,\"rowguid\":\"c0009006-715f-4b76-a05a-1a0d3adfb49a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":971}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1214.8500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-2000 Blue, 50\",\"ProductID\":971,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":35}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T44U-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":755.1508,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.42,\"rowguid\":\"84abda8f-0007-4bca-9a61-b2dea58866c3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":972}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1214.8500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-2000 Blue, 54\",\"ProductID\":972,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":35}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T44U-54\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"54\",\"StandardCost\":755.1508,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.68,\"rowguid\":\"6dcfe2a3-3555-44e4-9116-6f6dbe448b8b\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":973}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1700.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-350-W Yellow, 40\",\"ProductID\":973,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":27}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R79Y-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":1082.5100,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.35,\"rowguid\":\"237b16d9-53f2-4fd4-befe-48209e57aec3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":974}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1700.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-350-W Yellow, 42\",\"ProductID\":974,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":27}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R79Y-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":1082.5100,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":15.77,\"rowguid\":\"80bd3f8b-42c7-43d8-91f5-9fb6175287af\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":975}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1700.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-350-W Yellow, 44\",\"ProductID\":975,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":27}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R79Y-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":1082.5100,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":16.13,\"rowguid\":\"0c61e8af-003d-4e4b-b5b7-02f01a26be26\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":976}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Yellow\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":1700.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-350-W Yellow, 48\",\"ProductID\":976,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":27}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R79Y-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":1082.5100,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":16.42,\"rowguid\":\"ec4284dc-85fa-44a8-89ec-77fc9b71720a\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":977}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-750 Black, 58\",\"ProductID\":977,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":31}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R19B-58\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"58\",\"StandardCost\":343.6496,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.79,\"rowguid\":\"87b81a1a-a5b5-43d2-a20d-0230800490b9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":978}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Blue, 44\",\"ProductID\":978,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18U-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.77,\"rowguid\":\"4f638e15-2ed0-4193-90ce-47da580e01dd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":979}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Blue\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":742.3500,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Touring-3000 Blue, 50\",\"ProductID\":979,\"ProductLine\":\"T \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":36}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-T18U-50\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":3}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"50\",\"StandardCost\":461.4448,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":29.13,\"rowguid\":\"de305b62-88fc-465b-b23d-d8c0f7e6d361\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":980}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":769.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-400-W Silver, 38\",\"ProductID\":980,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":22}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M38S-38\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"38\",\"StandardCost\":419.7784,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":26.35,\"rowguid\":\"7a927632-99a4-4f24-adce-0062d2d113d9\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":981}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":769.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-400-W Silver, 40\",\"ProductID\":981,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":22}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M38S-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":419.7784,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":26.77,\"rowguid\":\"0a72791c-a984-4733-ae4e-2b4373cfd7cd\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":982}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":769.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-400-W Silver, 42\",\"ProductID\":982,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":22}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M38S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":419.7784,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.13,\"rowguid\":\"4ea4fe06-aaea-42d4-a9d9-69f6a9a4a042\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":983}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":769.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-400-W Silver, 46\",\"ProductID\":983,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":22}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M38S-46\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"46\",\"StandardCost\":419.7784,\"Style\":\"W \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.42,\"rowguid\":\"0b0c8ee4-ff2d-438e-9cac-464783b86191\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":984}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":564.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Silver, 40\",\"ProductID\":984,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18S-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":308.2179,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.35,\"rowguid\":\"b96c057b-6416-4851-8d59-bcb37c8e6e51\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":985}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":564.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Silver, 42\",\"ProductID\":985,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18S-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":308.2179,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.77,\"rowguid\":\"b8d1b5d9-8a39-4097-a04a-56e95559b534\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":986}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":564.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Silver, 44\",\"ProductID\":986,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18S-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":308.2179,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.13,\"rowguid\":\"e8cec794-e8e2-4312-96a7-4832e573d3fc\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":987}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":564.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Silver, 48\",\"ProductID\":987,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18S-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":308.2179,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.42,\"rowguid\":\"77ef419f-481f-40b9-bdb9-7e6678e550e3\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":988}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Silver\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":564.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Silver, 52\",\"ProductID\":988,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18S-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":308.2179,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.68,\"rowguid\":\"cad60945-183e-4ab3-a70c-3f5bac6bf134\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":989}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Black, 40\",\"ProductID\":989,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18B-40\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"40\",\"StandardCost\":294.5797,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.35,\"rowguid\":\"28287157-3f06-487b-8531-bee8a37329e4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":990}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Black, 42\",\"ProductID\":990,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18B-42\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"42\",\"StandardCost\":294.5797,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":27.77,\"rowguid\":\"f5752c9c-56ba-41a7-83fd-139da28c15fa\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":991}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Black, 44\",\"ProductID\":991,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":294.5797,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.13,\"rowguid\":\"c7852127-2fb8-4959-b5a3-de5a8d6445e4\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":992}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Black, 48\",\"ProductID\":992,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":294.5797,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.42,\"rowguid\":\"75752e26-a3b6-4264-9b06-f23a4fbdc5a7\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":993}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Mountain-500 Black, 52\",\"ProductID\":993,\"ProductLine\":\"M \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":23}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-M18B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":1}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":294.5797,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":28.68,\"rowguid\":\"69ee3b55-e142-4e4f-aed8-af02978fbe87\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":994}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":53.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"LL Bottom Bracket\",\"ProductID\":994,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":95}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BB-7421\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":5}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":23.9716,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":223.00,\"rowguid\":\"fa3c65cd-0a22-47e3-bdf6-53f1dc138c43\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":995}],\"EntitySetName\":\"Products\"},\"Class\":\"M \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":101.2400,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"ML Bottom Bracket\",\"ProductID\":995,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":96}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BB-8107\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":5}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":44.9506,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":168.00,\"rowguid\":\"71ab847f-d091-42d6-b735-7b0c2d82fc84\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":996}],\"EntitySetName\":\"Products\"},\"Class\":\"H \",\"Color\":null,\"DaysToManufacture\":1,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":121.4900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"HL Bottom Bracket\",\"ProductID\":996,\"ProductLine\":null,\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":97}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BB-9108\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":5}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":375,\"SafetyStockLevel\":500,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":null,\"StandardCost\":53.9416,\"Style\":null,\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"G  \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":null},\"Weight\":170.00,\"rowguid\":\"230c47c5-08b2-4ce3-b706-69c0bdd62965\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":997}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-750 Black, 44\",\"ProductID\":997,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":31}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R19B-44\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"44\",\"StandardCost\":343.6496,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":19.77,\"rowguid\":\"44ce4802-409f-43ab-9b27-ca53421805be\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":998}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-750 Black, 48\",\"ProductID\":998,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":31}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R19B-48\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"48\",\"StandardCost\":343.6496,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.13,\"rowguid\":\"3de9a212-1d49-40b6-b10a-f564d981dbde\"},{\"__type\":\"Product:#DataTests.AdventureWorks\",\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductID\",\"Value\":999}],\"EntitySetName\":\"Products\"},\"Class\":\"L \",\"Color\":\"Black\",\"DaysToManufacture\":4,\"DiscontinuedDate\":null,\"FinishedGoodsFlag\":true,\"ListPrice\":539.9900,\"MakeFlag\":true,\"ModifiedDate\":\"\\/Date(1079028096827-0800)\\/\",\"Name\":\"Road-750 Black, 52\",\"ProductID\":999,\"ProductLine\":\"R \",\"ProductModelReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductModelID\",\"Value\":31}],\"EntitySetName\":\"ProductModel\"}},\"ProductNumber\":\"BK-R19B-52\",\"ProductSubcategoryReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"ProductSubcategoryID\",\"Value\":2}],\"EntitySetName\":\"ProductSubcategory\"}},\"ReorderPoint\":75,\"SafetyStockLevel\":100,\"SellEndDate\":null,\"SellStartDate\":\"\\/Date(1057042800000-0700)\\/\",\"Size\":\"52\",\"StandardCost\":343.6496,\"Style\":\"U \",\"UnitMeasure1Reference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"LB \"}],\"EntitySetName\":\"UnitMeasure\"}},\"UnitMeasureReference\":{\"EntityKey\":{\"EntityContainerName\":\"AdventureWorksEntities\",\"EntityKeyValues\":[{\"Key\":\"UnitMeasureCode\",\"Value\":\"CM \"}],\"EntitySetName\":\"UnitMeasure\"}},\"Weight\":20.42,\"rowguid\":\"ae638923-2b67-4679-b90e-abbab17dca31\"}]}]";
        #endregion
    }
}
